extern crate proc_macro;
use case::CaseExt;

use proc_macro::TokenStream;
use quote::quote;
use syn::{parse_macro_input, Data, DeriveInput, Fields, Type, TypePath};

/// Get inner type. Example: Returns `T` for `Option<T>`.
fn extract_inner_type<'a>(ty: &'a Type, parent: &'a str) -> Option<&'a Type> {
    if let Type::Path(type_path) = ty {
        if type_path.path.segments.len() == 1 {
            let segment = &type_path.path.segments[0];
            if segment.ident == parent {
                if let syn::PathArguments::AngleBracketed(ref args) = segment.arguments {
                    if args.args.len() == 1 {
                        if let syn::GenericArgument::Type(ref inner_type) = args.args[0] {
                            return Some(inner_type);
                        }
                    }
                }
            }
        }
    }
    None
}

fn type_path(ty: &syn::Type) -> TypePath {
    if let syn::Type::Path(type_path) = ty {
        return type_path.clone();
    }
    unreachable!();
}

fn is_type(ty: &syn::Type, t: &str) -> bool {
    let type_path = type_path(ty);
    let id = type_path.path.segments.first().unwrap().ident.to_string();
    id == t
}

#[proc_macro_derive(Model)]
pub fn model_derive(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident;

    // Generate code for each field
    let field_code = if let Data::Struct(data_struct) = input.data {
        match data_struct.fields {
            Fields::Named(fields_named) => {
                let field_process_code: Vec<_> = fields_named.named.iter().map(|field| {
                    let field_name = &field.ident.as_ref().unwrap();
                    let field_type = &field.ty;
                    let field_name_str = field_name.to_string();

                    if field_name_str == "_id" {
                        return quote! {};
                    }

                    if is_type(field_type, "Historic") {
                        let inner_field_type = extract_inner_type(field_type, "Historic").unwrap();
                        if is_type(inner_field_type, "Vec") {
                            return quote! {
                                mongod::update_historic_vec!(self, obj, #field_name_str, #field_name, update);
                            };
                        }

                        return quote! {
                            mongod::update_historic_str!(self, obj, #field_name_str, #field_name, update);
                        }
                    }

                    if is_type(field_type, "Option") {
                        let inner_field_type = extract_inner_type(field_type, "Option").unwrap();

                        if is_type(inner_field_type, "Historic") {
                            let sub_inner_field_type = extract_inner_type(inner_field_type, "Historic").unwrap();
                            if is_type(sub_inner_field_type, "Reference") {
                                return quote! {
                                    mongod::update_historic_ref_option!(self, obj, #field_name_str, #field_name, update);
                                };
                            }
                        }

                        return quote! {
                            mongod::update_value_option!(self, obj, #field_name_str, #field_name, update, #inner_field_type);
                        };
                    }

                    quote! {
                        mongod::update_value!(self, obj, #field_name_str, #field_name, update, #field_type);
                    }
                }).collect();

                quote! {
                    impl mongod::model::Model for #name {
                        async fn update_values(
                            &mut self,
                            obj: &serde_json::Map<String, serde_json::Value>,
                            update: &mut mongodb::bson::Document,
                        ) {
                            #( #field_process_code )*
                        }
                    }
                }
            }
            _ => unimplemented!(),
        }
    } else {
        unimplemented!()
    };

    TokenStream::from(field_code)
}

#[proc_macro_derive(Referencable)]
pub fn referencable_derive(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident;
    let name_str = name.to_string().to_snake();

    let code = quote! {
        impl mongod::model::reference::Referencable for #name {
            fn collection_name() -> &'static str {
                #name_str
            }

            fn id(&self) -> &str {
                &self._id
            }
        }
    };

    TokenStream::from(code)
}
